//
// Author:
//       smdn <smdn@smdn.jp>
//
// Copyright (c) 2008-2014 smdn
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#if NET_4_0
#define ASYNC_OPS
#endif

using System;
using System.Collections.Generic;
using System.IO;
#if ASYNC_OPS
using System.Threading;
using System.Threading.Tasks;
#endif

using Smdn.IO;
using Smdn.Net.Imap4.Client.Session;
using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.Client;

namespace Smdn.Net.Imap4.Client {
  partial class ImapMailboxInfo {
    /*
     * APPEND
     */
    public void AppendMessage(Stream stream)
    {
      AppendMessage(new ImapAppendMessage(stream));
    }

    public void AppendMessage(Stream stream,
                              DateTimeOffset appendDate)
    {
      AppendMessage(new ImapAppendMessage(stream,
                                          appendDate));
    }

    public void AppendMessage(Stream stream,
                              IImapMessageFlagSet flags)
    {
      AppendMessage(new ImapAppendMessage(stream,
                                          flags));
    }

    public void AppendMessage(Stream stream,
                              DateTimeOffset appendDate,
                              IImapMessageFlagSet flags)
    {
      AppendMessage(new ImapAppendMessage(stream,
                                          appendDate,
                                          flags));
    }

    public void AppendMessage(IImapAppendMessage message)
    {
      if (message == null)
        throw new ArgumentNullException("message");

      ThrowIfNonExistentOrUnselectable();

      client.RunExclusively(() => {
        ProcessResult(Client.Session.Append(message, Mailbox),
                      ThrowIfAppendErrorResponse);
      });
    }

#if ASYNC_OPS
    public Task AppendMessageAsync(Stream stream)
    {
      return AppendMessageAsync(new ImapAppendMessage(stream));
    }

    public Task AppendMessageAsync(Stream stream,
                                   DateTimeOffset appendDate)
    {
      return AppendMessageAsync(new ImapAppendMessage(stream,
                                                      appendDate));
    }

    public Task AppendMessageAsync(Stream stream,
                                   IImapMessageFlagSet flags)
    {
      return AppendMessageAsync(new ImapAppendMessage(stream,
                                                      flags));
    }

    public Task AppendMessageAsync(Stream stream,
                                   DateTimeOffset appendDate,
                                   IImapMessageFlagSet flags)
    {
      return AppendMessageAsync(new ImapAppendMessage(stream,
                                                      appendDate,
                                                      flags));
    }

    public Task AppendMessageAsync(IImapAppendMessage message)
    {
      if (message == null)
        throw new ArgumentNullException("message");

      ThrowIfNonExistentOrUnselectable();

      return client.StartTaskExclusively(() => {
        ProcessResult(Client.Session.Append(message, Mailbox),
                      ThrowIfAppendErrorResponse);
      }, CancellationToken.None);
    }
#endif

    public void AppendMessages(IEnumerable<IImapAppendMessage> messages)
    {
      if (messages == null)
        throw new ArgumentNullException("messages");

      ThrowIfNonExistentOrUnselectable();

      client.RunExclusively(() => AppendMessagesCore(messages));
    }

#if ASYNC_OPS
    public Task AppendMessagesAsync(IEnumerable<IImapAppendMessage> messages)
    {
      if (messages == null)
        throw new ArgumentNullException("messages");

      ThrowIfNonExistentOrUnselectable();

      return client.StartTaskExclusively(() => AppendMessagesCore(messages),
                                         CancellationToken.None);
    }
#endif

    private void AppendMessagesCore(IEnumerable<IImapAppendMessage> messages)
    {
      if (Client.IsCapabilityAvailable(ImapCapability.MultiAppend)) {
        ProcessResult(Client.Session.AppendMultiple(messages, Mailbox),
                      ThrowIfAppendErrorResponse);
      }
      else {
        foreach (var message in messages) {
          ProcessResult(Client.Session.Append(message, Mailbox),
                        ThrowIfAppendErrorResponse);
        }
      }
    }

    public void AppendMessage(Action<Stream> writeMessage)
    {
      AppendMessageWithWriteAction(null,
                                   null,
                                   null,
                                   writeMessage);
    }

    public void AppendMessage(long length,
                              Action<Stream> writeMessage)
    {
      AppendMessageWithWriteAction(length,
                                   null,
                                   null,
                                   writeMessage);
    }

    public void AppendMessage(DateTimeOffset appendDate,
                              Action<Stream> writeMessage)
    {
      AppendMessageWithWriteAction(null,
                                   appendDate,
                                   null,
                                   writeMessage);
    }

    public void AppendMessage(IImapMessageFlagSet flags,
                              Action<Stream> writeMessage)
    {
      AppendMessageWithWriteAction(null,
                                   null,
                                   flags,
                                   writeMessage);
    }

    public void AppendMessage(DateTimeOffset appendDate,
                              IImapMessageFlagSet flags,
                              Action<Stream> writeMessage)
    {
      AppendMessageWithWriteAction(null,
                                   appendDate,
                                   flags,
                                   writeMessage);
    }

    public void AppendMessage(long length,
                              DateTimeOffset appendDate,
                              Action<Stream> writeMessage)
    {
      AppendMessageWithWriteAction(length,
                                   appendDate,
                                   null,
                                   writeMessage);
    }

    public void AppendMessage(long length,
                              IImapMessageFlagSet flags,
                              Action<Stream> writeMessage)
    {
      AppendMessageWithWriteAction(length,
                                   null,
                                   flags,
                                   writeMessage);
    }

    public void AppendMessage(long length,
                              DateTimeOffset appendDate,
                              IImapMessageFlagSet flags,
                              Action<Stream> writeMessage)
    {
      AppendMessageWithWriteAction(length,
                                   appendDate,
                                   flags,
                                   writeMessage);
    }

#if ASYNC_OPS
    public Task AppendMessageAsync(Action<Stream> writeMessage)
    {
      return AppendMessageWithWriteActionAsync(null,
                                               null,
                                               null,
                                               writeMessage);
    }

    public Task AppendMessageAsync(long length,
                                   Action<Stream> writeMessage)
    {
      return AppendMessageWithWriteActionAsync(length,
                                               null,
                                               null,
                                               writeMessage);
    }

    public Task AppendMessageAsync(DateTimeOffset appendDate,
                                   Action<Stream> writeMessage)
    {
      return AppendMessageWithWriteActionAsync(null,
                                               appendDate,
                                               null,
                                               writeMessage);
    }

    public Task AppendMessageAsync(IImapMessageFlagSet flags,
                                   Action<Stream> writeMessage)
    {
      return AppendMessageWithWriteActionAsync(null,
                                               null,
                                               flags,
                                               writeMessage);
    }

    public Task AppendMessageAsync(DateTimeOffset appendDate,
                                   IImapMessageFlagSet flags,
                                   Action<Stream> writeMessage)
    {
      return AppendMessageWithWriteActionAsync(null,
                                               appendDate,
                                               flags,
                                               writeMessage);
    }

    public Task AppendMessageAsync(long length,
                                   DateTimeOffset appendDate,
                                   Action<Stream> writeMessage)
    {
      return AppendMessageWithWriteActionAsync(length,
                                               appendDate,
                                               null,
                                               writeMessage);
    }

    public Task AppendMessageAsync(long length,
                                   IImapMessageFlagSet flags,
                                   Action<Stream> writeMessage)
    {
      return AppendMessageWithWriteActionAsync(length,
                                               null,
                                               flags,
                                               writeMessage);
    }

    public Task AppendMessageAsync(long length,
                                   DateTimeOffset appendDate,
                                   IImapMessageFlagSet flags,
                                   Action<Stream> writeMessage)
    {
      return AppendMessageWithWriteActionAsync(length,
                                               appendDate,
                                               flags,
                                               writeMessage);
    }
#endif

    private void AppendMessageWithWriteAction(long? length,
                                              DateTimeOffset? appendDate,
                                              IImapMessageFlagSet flags,
                                              Action<Stream> writeMessage)
    {
      if (writeMessage == null)
        throw new ArgumentNullException("writeMessage");
      if (length.HasValue && length.Value < 0L)
        throw ExceptionUtils.CreateArgumentMustBeZeroOrPositive("length", length.Value);

      ThrowIfNonExistentOrUnselectable();

      client.RunExclusively(() => AppendMessageCore(length,
                                                    appendDate,
                                                    flags,
                                                    writeMessage));
    }

#if ASYNC_OPS
    private Task AppendMessageWithWriteActionAsync(long? length,
                                                   DateTimeOffset? appendDate,
                                                   IImapMessageFlagSet flags,
                                                   Action<Stream> writeMessage)
    {
      if (writeMessage == null)
        throw new ArgumentNullException("writeMessage");
      if (length.HasValue && length.Value < 0L)
        throw ExceptionUtils.CreateArgumentMustBeZeroOrPositive("length", length.Value);

      ThrowIfNonExistentOrUnselectable();

      return client.StartTaskExclusively(() => AppendMessageCore(length,
                                                                 appendDate,
                                                                 flags,
                                                                 writeMessage),
                                         CancellationToken.None);
    }
#endif

    public void AppendMessage(ImapMessageInfo message)
    {
      if (message == null)
        throw new ArgumentNullException("message");

      if (message.Client == this.client) {
        message.CopyTo(this);
      }
      else {
        ThrowIfNonExistentOrUnselectable();

        client.RunExclusively(() => CopyMessageAcrossSessionCore(message, true));
      }
    }

#if ASYNC_OPS
    public Task AppendMessageAsync(ImapMessageInfo message)
    {
      if (message == null)
        throw new ArgumentNullException("message");

      if (message.Client == this.client) {
        return message.CopyToAsync(this);
      }
      else {
        ThrowIfNonExistentOrUnselectable();

        return client.StartTaskExclusively(() => CopyMessageAcrossSessionCore(message, true),
                                           CancellationToken.None);
      }
    }
#endif

    private static readonly ImapFetchDataItem appendingDataItem =
      ImapFetchDataItem.Rfc822Size +
      ImapFetchDataItem.InternalDate +
      ImapFetchDataItem.Flags;

    internal void CopyMessageAcrossSession(ImapMessageInfo message, bool needToSetMessageBusy)
    {
#if DEBUG
      if (message.Client == this.client)
        throw new NotSupportedException("use ImapMessageInfo.CopyTo");
#endif

      client.RunExclusively(() => CopyMessageAcrossSessionCore(message, needToSetMessageBusy));
    }

    private void CopyMessageAcrossSessionCore(ImapMessageInfo message, bool needToSetMessageBusy)
    {
      IImapMessageAttribute attr;

      using (var sourceStream = message.OpenReadWithAttributes(appendingDataItem,
                                                               needToSetMessageBusy,
                                                               out attr)) {
        AppendMessageCore(attr.Rfc822Size,
                          attr.InternalDate,
                          attr.Flags,
                          delegate(Stream destStream) {
          sourceStream.CopyTo(destStream,
                              ImapFetchMessageBodyStream.DefaultFetchBlockSize);
        });
      }
    }

    private void AppendMessageCore(long? length,
                                   DateTimeOffset? appendDate,
                                   IImapMessageFlagSet flags,
                                   Action<Stream> writeMessage)
    {
      var appendContext = client.Session.PrepareAppend(length,
                                                       appendDate,
                                                       flags,
                                                       mailbox);

      writeMessage(appendContext.WriteStream);

      ImapAppendedUidSet appendedUidSet;

      ProcessResult(appendContext.GetResult(out appendedUidSet),
                    ThrowIfAppendErrorResponse);
    }

    private bool ThrowIfAppendErrorResponse(ImapTaggedStatusResponse status)
    {
      if (status.ResponseText.Code == ImapResponseCode.TryCreate)
        throw new ImapMailboxNotFoundException(FullName);
      else
        return true;
    }
  }
}

